#include "Grid.h"
#include "Action.h"
#include "Diff.h"
#include "Pos.h"
#include <map>
#include <set>

using std::map;
using std::set;

namespace
{
    void onEmpty(Grid& grid, Pos pos)
    {
        GridInfo& info = *grid.m_pInfo;
        uint32_t x = pos.m_x;
        uint32_t y = pos.m_y;

        {
            CellState cs = grid.cell(x-1, y-1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x-1, y-1)); }
        }

        {
            CellState cs = grid.cell(x, y-1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x, y-1)); }
        }

        {
            CellState cs = grid.cell(x+1, y-1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x+1, y-1)); }
        }

        {
            CellState cs = grid.cell(x-1, y);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x-1, y)); }
            if (csIsRock(cs)) { info.m_fallingRocks.insert(Pos(x-1, y)); }
        }

        {
            CellState cs = grid.cell(x+1, y);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x+1, y)); }
            if (csIsRock(cs)) { info.m_fallingRocks.insert(Pos(x+1, y)); }
        }

        {
            CellState cs = grid.cell(x-1, y+1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x-1, y+1)); }
            if (csIsRock(cs)) { info.m_fallingRocks.insert(Pos(x-1, y+1)); }
        }

        {
            CellState cs = grid.cell(x, y+1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x, y+1)); }
            if (csIsRock(cs)) { info.m_fallingRocks.insert(Pos(x, y+1)); }
        }

        {
            CellState cs = grid.cell(x+1, y+1);
            if (csIsBeard(cs)) { info.m_growingBeards.insert(Pos(x+1, y+1)); }
            if (csIsRock(cs)) { info.m_fallingRocks.insert(Pos(x+1, y+1)); }
        }
    }
}

void step(Grid *pGrid, Action ac, Diff* pDiff, bool* pStupid)
{
    bool stupidDummy = false;
    if (!pStupid) { pStupid = &stupidDummy; }
    //bool gotPoints = false;
    bool waitUseful = false;

    Grid& grid = *pGrid;
    GridInfo& info = *grid.m_pInfo;
    Diff& diff = *pDiff;

    diff.m_pInfo = new GridInfo(info);
    //diff.m_cells.clear();

    map<Pos, CellState> cells;
    set<Pos> cellsInDiff;

    if (info.m_state != State_Running) {
        *pStupid = true;
        return;
    }

    // Move
    if (acIsAbort(ac)) {
        //info.m_score += 25 * info.m_lambdaCollected;
        info.m_state = State_Aborted;
        return;
    }

    info.m_time++;
    info.m_score -= 1;

    Pos curPos = info.m_robotPos;
    Pos newPos = curPos;
    Pos pushPos = curPos;

    bool isMove = false;
    bool isHorizontal = false;
    if (acIsMoveUp(ac)) {
        newPos.m_y = curPos.m_y + 1;
        isMove = true;
    } else if (acIsMoveDown(ac)) {
        newPos.m_y = curPos.m_y - 1;
        isMove = true;
    } else if (acIsMoveLeft(ac)) {
        newPos.m_x = curPos.m_x - 1;
        pushPos.m_x = curPos.m_x - 2;
        isMove = true;
        isHorizontal = true;
    } else if (acIsMoveRight(ac)) {
        newPos.m_x = curPos.m_x + 1;
        pushPos.m_x = curPos.m_x + 2;
        isMove = true;
        isHorizontal = true;
    }

    if (isMove) {
        CellState csNew = grid.cell(newPos);
        if (csIsEmpty(csNew) ||
            csIsEarth(csNew) ||
            csIsRazor(csNew) ||
            csIsLambda(csNew) ||
            (csIsLift(csNew) &&
             info.m_lambdaCollected == grid.m_lambdaTotal)) {

            if (csIsLambda(csNew)) {
                info.m_lambdaCollected++;
                info.m_score += 50;
                //gotPoints = true;
            }

            if (csIsLift(csNew)) {
                info.m_score += 25 * info.m_lambdaCollected;
                info.m_state = State_Finished;
                //gotPoints = true;
            }

            if (csIsRazor(csNew)) {
                info.m_razors++;
            }

            diff.m_cells.push_back(CellDiff(curPos, CS_Robot));
            diff.m_cells.push_back(CellDiff(newPos, csNew));
            cellsInDiff.insert(curPos);
            cellsInDiff.insert(newPos);
            grid.cell(curPos) = CS_Empty;
            grid.cell(newPos) = CS_Robot;
            info.m_robotPos = newPos;
            onEmpty(grid, curPos);
        } else if (csIsRock(csNew) && isHorizontal) {
            CellState csPush = grid.cell(pushPos);
            if (csIsEmpty(csPush)) {
                diff.m_cells.push_back(CellDiff(curPos, CS_Robot));
                diff.m_cells.push_back(CellDiff(newPos, csNew));
                diff.m_cells.push_back(CellDiff(pushPos, csPush));
                cellsInDiff.insert(curPos);
                cellsInDiff.insert(newPos);
                cellsInDiff.insert(pushPos);
                grid.cell(curPos) = CS_Empty;
                grid.cell(newPos) = CS_Robot;
                grid.cell(pushPos) = csNew;
                info.m_robotPos = newPos;
                onEmpty(grid, curPos);
                info.m_fallingRocks.insert(pushPos);
            } else {
                *pStupid = true;
            }
        } else if (csIsTrampoline(csNew)) {
            const Trampoline& trampoline =
                grid.m_trampolines[csGetNumber(csNew)-1];
            const Target& target =
                grid.m_targets[trampoline.m_target-1];

            diff.m_cells.push_back(CellDiff(curPos, CS_Robot));
            cellsInDiff.insert(curPos);
            grid.cell(curPos) = CS_Empty;
            onEmpty(grid, curPos);

            for (uint32_t i = 1; i <= 9; i++) {
                const Trampoline& curTrampoline = grid.m_trampolines[i-1];
                if (curTrampoline.m_target == trampoline.m_target) {
                    diff.m_cells.push_back(CellDiff(curTrampoline.m_pos,
                                                    grid.cell(curTrampoline.m_pos)));
                    cellsInDiff.insert(curTrampoline.m_pos);
                    grid.cell(curTrampoline.m_pos) = CS_Empty;
                    onEmpty(grid, curTrampoline.m_pos);
                }
            }


            diff.m_cells.push_back(CellDiff(target.m_pos,
                                            grid.cell(target.m_pos)));
            cellsInDiff.insert(target.m_pos);
            grid.cell(target.m_pos) = CS_Robot;

            info.m_robotPos = target.m_pos;
        } else {
            *pStupid = true;
        }
    }

    if (acIsShave(ac)) {
        if (info.m_razors == 0) {
            *pStupid = true;
        } else {
            info.m_razors--;
            bool shaved = false;
            for (uint32_t y = curPos.m_y-1; y <= curPos.m_y+1; y++) {
                for (uint32_t x = curPos.m_x-1; x <= curPos.m_x+1; x++) {
                    if (csIsBeard(grid.cell(x, y))) {

                        diff.m_cells.push_back(CellDiff(Pos(x, y),
                                                        CS_Beard));
                        cellsInDiff.insert(Pos(x, y));
                        grid.cell(x, y) = CS_Empty;
                        onEmpty(grid, Pos(x, y));
                        shaved = true;
                    }
                }
            }
            if (!shaved) {
                *pStupid = true;
            }
        }
    }

    // Map update
    set<Pos> oldFallingRocks;
    oldFallingRocks.swap(info.m_fallingRocks);
    set<Pos> oldGrowingBeards;
    if (grid.m_growth != 0 &&
        !info.m_growingBeards.empty()) {
        waitUseful = true;
    }
    if (grid.m_growth != 0 &&
        info.m_time % grid.m_growth == 0) {
        oldGrowingBeards.swap(info.m_growingBeards);
    }
    set<Pos> newSpaces;

    {
        set<Pos>::const_iterator rockIter = oldFallingRocks.begin();
        set<Pos>::const_iterator rockEnd = oldFallingRocks.end();
        set<Pos>::const_iterator beardIter = oldGrowingBeards.begin();
        set<Pos>::const_iterator beardEnd = oldGrowingBeards.end();
        while (rockIter != rockEnd || beardIter != beardEnd) {
            if (beardIter == beardEnd ||
                (rockIter != rockEnd &&
                 *rockIter < *beardIter)) {

                const Pos& pos = *rockIter;
                ++rockIter;
                uint32_t x = pos.m_x;
                uint32_t y = pos.m_y;

                CellState cs = grid.cell(x, y);
                if (csIsRock(cs)) {
                    CellState csUnder = grid.cell(x, y-1);
                    bool fall = false;
                    uint32_t fallX = 0;
                    uint32_t fallY = 0;
                    if (csIsEmpty(csUnder)) {
                        fall = true;
                        fallX = x;
                        fallY = y-1;
                    } else if (csIsRock(csUnder)) {
                        if (csIsEmpty(grid.cell(x+1, y)) &&
                            csIsEmpty(grid.cell(x+1, y-1))) {
                            fall = true;
                            fallX = x+1;
                            fallY = y-1;
                        } else if (csIsEmpty(grid.cell(x-1, y)) &&
                                   csIsEmpty(grid.cell(x-1, y-1))) {
                            fall = true;
                            fallX = x-1;
                            fallY = y-1;
                        }
                    } else if (csIsLambda(csUnder)) {
                        if (csIsEmpty(grid.cell(x+1, y)) &&
                            csIsEmpty(grid.cell(x+1, y-1))) {
                            fall = true;
                            fallX = x+1;
                            fallY = y-1;
                        }
                    }
                    if (fall) {
                        bool shatter =
                            csIsHORock(cs) &&
                            !csIsEmpty(grid.cell(fallX, fallY-1));
                        cells[Pos(x, y)] = CS_Empty;
                        cells[Pos(fallX, fallY)] = shatter ? CS_Lambda : cs;

                        if (info.m_robotPos.m_x == fallX &&
                            info.m_robotPos.m_y == fallY - 1 &&
                            info.m_state == State_Running) {
                            info.m_state = State_Broken;
                        }

                        newSpaces.insert(Pos(x, y));
                        if (!shatter) {
                            info.m_fallingRocks.insert(Pos(fallX, fallY));
                        }
                        waitUseful = true;
                    }
                }
            } else {
                const Pos& pos = *beardIter;
                ++beardIter;

                if (csIsBeard(grid.cell(pos))) {
                    for (uint32_t y = pos.m_y-1; y <= pos.m_y+1; y++) {
                        for (uint32_t x = pos.m_x-1; x <= pos.m_x+1; x++) {
                            if (csIsEmpty(grid.cell(x, y))) {
                                cells[Pos(x, y)] = CS_Beard;
                                info.m_growingBeards.insert(Pos(x, y));
                            }
                        }
                    }
                }
            }
        }
    }

    {
        map<Pos, CellState>::const_iterator iter = cells.begin();
        map<Pos, CellState>::const_iterator end = cells.end();
        for (; iter != end; ++iter) {
            const Pos& pos = iter->first;
            if (cellsInDiff.find(pos) == cellsInDiff.end()) {
                diff.m_cells.push_back(CellDiff(pos, grid.cell(pos)));
            }
            grid.cell(pos) = iter->second;
        }
    }

    {
        set<Pos>::const_iterator iter = newSpaces.begin();
        set<Pos>::const_iterator end = newSpaces.end();
        for (; iter != end; ++iter) {
            const Pos& pos = *iter;
            onEmpty(grid, pos);
        }
    }

    if (grid.m_flooding != 0 &&
        info.m_time % grid.m_flooding == 0 &&
        info.m_water < grid.m_height) {
        info.m_water++;
    }

    if (info.m_robotPos.m_y > info.m_water) {
        info.m_underwater = 0;
    } else {
        info.m_underwater++;
        if (info.m_underwater > grid.m_waterproof) {
            info.m_state = State_Broken;
        }
    }

    if (info.m_state == State_Broken) {
        info.m_score -= 25 * info.m_lambdaCollected;
        if (info.m_score <= diff.m_pInfo->m_score) {
            *pStupid = true;
        }
    }

    if (acIsWait(ac) && !waitUseful) {
        *pStupid = true;
    }
}
