#include "Solve.h"
#include "Grid.h"
#include "Sim.h"
#include "Diff.h"
#include "SignalHandler.h"
#include "Zone.h"
#include <set>
#include <map>
#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

using std::set;
using std::map;
using std::vector;

Action actionsToTry[] = {
    Action_MoveUp,
    Action_MoveDown,
    Action_MoveLeft,
    Action_MoveRight,
    Action_Shave,
    Action_Wait
};
const uint32_t numActionsToTry = sizeof(actionsToTry)/sizeof(actionsToTry[0]);

static uint32_t s_nodeCount = 0;

class Node
{
public:
    Node();
    ~Node();

    void addRef() { m_refCount++; }
    void release() { m_refCount--; if (m_refCount == 0) { delete this; } }

    void setParent(Node* parent);
    void setDiffParent(Node* diffParent);

    Node* m_parent;
    Node* m_diffParent;
    Action m_action;
    Diff m_diff;
    uint32_t m_refCount;
};

Node::Node() :
    m_parent(0),
    m_diffParent(0),
    m_action(Action_Wait),
    m_diff(),
    m_refCount(1)
{
    s_nodeCount++;
}

Node::~Node()
{
    if (m_parent) { m_parent->release(); }
    if (m_diffParent) { m_diffParent->release(); }
    s_nodeCount--;
}

void Node::setParent(Node* parent)
{
    if (parent) { parent->addRef(); }
    if (m_parent) { m_parent->release(); }
    m_parent = parent;
}

void Node::setDiffParent(Node* diffParent)
{
    if (diffParent) { diffParent->addRef(); }
    if (m_diffParent) { m_diffParent->release(); }
    m_diffParent = diffParent;
}

class Best
{
public:
    Best();
    ~Best();

    void update(const Grid& grid, Node* pNode);
    void clear();

    Node* m_pNode;
    int32_t m_score;
    State m_state;
};

Best::Best() :
    m_pNode(0),
    m_score(0x80000000),
    m_state(State_Running)
{
}

Best::~Best()
{
}

void Best::update(const Grid& grid, Node* pNode)
{
    if (grid.m_pInfo->m_score > m_score) {
        if (pNode) { pNode->addRef(); }
        if (m_pNode) { m_pNode->release(); }
        m_pNode = pNode;
        m_score = grid.m_pInfo->m_score;
        m_state = grid.m_pInfo->m_state;
    }
}

void Best::clear()
{
    if (m_pNode) { m_pNode->release(); }
    m_pNode = 0;
    m_score = 0x80000000;
    m_state = State_Running;
}

class Bucket
{
public:
    Bucket(Pos robotPos,
           bool justShaved,
           bool justWaited) :
        m_robotPos(robotPos),
        m_justShaved(justShaved),
        m_justWaited(justWaited) {}
    ~Bucket() {}

    bool operator<(const Bucket& other) const {
        if (m_robotPos != other.m_robotPos) {
            return m_robotPos < other.m_robotPos;
        }

        if (m_justShaved != other.m_justShaved) {
            return m_justShaved < other.m_justShaved;
        }

        if (m_justWaited != other.m_justWaited) {
            return m_justWaited < other.m_justWaited;
        }

        return false;
    }

    Pos m_robotPos;
    bool m_justShaved;
    bool m_justWaited;
};

class BucketMap
{
public:
    BucketMap();
    ~BucketMap();

    Node* getNode(const Bucket& bucket) const;
    void setNode(const Bucket& bucket, Node* pNode);
    void clear();

    map<Bucket, Node*> m_map;
};

BucketMap::BucketMap() :
    m_map()
{
}

BucketMap::~BucketMap()
{
}

Node* BucketMap::getNode(const Bucket& bucket) const
{
    map<Bucket, Node*>::const_iterator findIter =
        m_map.find(bucket);
    if (findIter != m_map.end()) {
        return findIter->second;
    }

    return 0;
}

void BucketMap::setNode(const Bucket& bucket, Node* pNode)
{
    Node*& pBucketNode = m_map[bucket];
    if (pNode) { pNode->addRef(); }
    if (pBucketNode) { pBucketNode->release(); }
    pBucketNode = pNode;
}

void BucketMap::clear()
{
    map<Bucket, Node*>::iterator iter = m_map.begin();
    map<Bucket, Node*>::iterator end = m_map.end();
    for (; iter != end; ++iter) {
        Node* pNode = iter->second;
        if (pNode) { pNode->release(); }
    }
}

class Plan
{
public:
    Plan();
    ~Plan();

    bool contains(const Grid& grid,
                  const Diff& Diff) const;

    set<Pos> m_set;
    set<Pos> m_visited;
    Pos m_lastVisited;
    vector<Pos> m_bestVisited;
    map<Pos, Pos> m_parentMap;
    map<Pos, Pos> m_nextMap;
};

Plan::Plan()
{
}

Plan::~Plan()
{
}

bool Plan::contains(const Grid& grid,
                    const Diff& diff) const
{
    Pos robotPos = grid.m_pInfo->m_robotPos;
    if (m_visited.find(robotPos) != m_visited.end()) {
        return false;
    }

    map<Pos, Pos>::const_iterator findIter = m_parentMap.find(robotPos);
    if (findIter != m_parentMap.end()) {
        if (m_visited.find(findIter->second) == m_visited.end()) {
            return false;
        }
    }

    if (m_lastVisited != Pos()) {
        map<Pos, Pos>::const_iterator findIter = m_nextMap.find(m_lastVisited);
        if (findIter != m_nextMap.end()) {
            Pos nextPos = findIter->second;
            if (m_visited.find(nextPos) == m_visited.end() &&
                robotPos != nextPos) {
                return false;
            }
        }
    }

    if (grid.m_pInfo->m_state == State_Finished) {
        return true;
    }

    if (grid.m_pInfo->m_lambdaCollected >
        diff.m_pInfo->m_lambdaCollected) {
        return true;
    }

    if (m_set.find(robotPos) != m_set.end()) {
        return true;
    }

    return false;
}

void search(Grid& grid, Node *pStartNode, Best& best,
            Plan& plan,
            vector<Node*>* pNodes,
            bool bTest)
{
    bool done = false;

    pStartNode->addRef();

    BucketMap bucketMap;

    vector<Node*> oldNodes;
    vector<Node*> newNodes;

    oldNodes.push_back(pStartNode);
    pStartNode->addRef();

    vector<Node*> path;

    if (bTest) {
        //puts("\x1b[2J\x1b[H");
        //writeGrid(stdout, grid, true);
    }

    bool first = true;
    while (!oldNodes.empty()) {
        if (receivedSignal()) { break; }
        if (done) { break; }

        for (size_t iNode = 0; iNode < oldNodes.size(); iNode++) {
            if (receivedSignal()) { break; }
            size_t firstNewNode = newNodes.size();
            Node* pNode = oldNodes[iNode];

            Node* pCurNode = pNode;
            while (pCurNode->m_diffParent) {
                path.push_back(pCurNode);
                pCurNode = pCurNode->m_diffParent;
            }

            for (size_t iPath = path.size(); iPath > 0; iPath--) {
                applyDiff(&grid, &path[iPath-1]->m_diff);
            }

            if (first) {
                Bucket startBucket(grid.m_pInfo->m_robotPos,
                                   false,
                                   false);
                bucketMap.setNode(startBucket, pStartNode);
            }

            uint32_t curTime = grid.m_pInfo->m_time;
            uint32_t parentTime = curTime & (curTime - 1);
            //uint32_t parentTime = 0;

            for (uint32_t iAction = 0; iAction < numActionsToTry; iAction++) {
                if (receivedSignal()) { break; }
                if (done) { break; }

                if (grid.m_pInfo->m_state != State_Running) {
                    continue;
                }

                Node* pNewNode = new Node;
                Action action = actionsToTry[iAction];
                Diff& diff = pNewNode->m_diff;
                bool stupid = false;
                step(&grid, action, &diff, &stupid);

                if (!stupid) {
                    Bucket bucket(grid.m_pInfo->m_robotPos,
                                  acIsShave(action),
                                  acIsWait(action));
                    if (bucketMap.getNode(bucket) == 0) {
                        pNewNode->setParent(pNode);
                        pNewNode->m_action = action;
                        pNewNode->setDiffParent(pNode);
                        newNodes.push_back(pNewNode);
                        pNewNode->addRef();

                        bucketMap.setNode(bucket, pNewNode);

                        if (plan.contains(grid, diff)) {
                            if (pNodes) {
                                if (pNodes->size() < 1) {
                                    pNewNode->addRef();
                                    pNodes->push_back(pNewNode);
                                    plan.m_visited.insert(grid.m_pInfo->m_robotPos);
                                    plan.m_lastVisited = grid.m_pInfo->m_robotPos;

                                    if (bTest) {
                                        //puts("\x1b[H");
                                        //writeGrid(stdout, grid, true);
                                    }

                                    if (pNodes->size() == 1) {
                                        done = true;
                                    }
                                }
                            }
                        }

                        if (bTest) {
                            //puts("\x1b[H");
                            //writeGrid(stdout, grid, true);
                        }

                        int32_t oldScore = best.m_score;
                        best.update(grid, pNewNode);
                        if (best.m_score > oldScore) {
                            if (bTest) {
                                puts("\x1b[H");
                                writeGrid(stdout, grid, true);
                                //printf("Best score: %d\n", best.m_score);
                            }
                        }
                    }
                }

                applyDiff(&grid, &diff);
                pNewNode->release();
            }

            for (size_t iPath = 0; iPath < path.size(); iPath++) {
                applyDiff(&grid, &path[iPath]->m_diff);

                if (grid.m_pInfo->m_time >= parentTime) {
                    for (size_t iNewNode = firstNewNode;
                         iNewNode < newNodes.size(); iNewNode++) {
                        //if (receivedSignal()) { break; }
                        concatDiff(grid,
                                   path[iPath]->m_diff,
                                   &newNodes[iNewNode]->m_diff);
                        newNodes[iNewNode]->setDiffParent(
                            path[iPath]->m_diffParent);
                    }
                }
            }
            path.clear();
        }

        for (size_t iNode = 0; iNode < oldNodes.size(); iNode++) {
            oldNodes[iNode]->release();
        }
        oldNodes.clear();
        oldNodes.swap(newNodes);
        first = false;
    }

    for (size_t iNode = 0; iNode < oldNodes.size(); iNode++) {
        oldNodes[iNode]->release();
    }
    oldNodes.clear();

    pStartNode->release();

    bucketMap.clear();
}

int32_t tryPlan(Grid& grid, Node* pRootNode, Best& best, Plan& plan, bool bTest)
{
    plan.m_visited.clear();

    Node* pNode = pRootNode;
    pNode->addRef();

    bool done = false;
    do {
        if (receivedSignal()) { break; }
        vector<Node*> newNodes;
        search(grid, pNode, best, plan, &newNodes, bTest);
        if (newNodes.empty()) {
            done = true;
        } else {
            Node *pNewNode = newNodes[0];
            pNewNode->addRef();
            pNode->release();
            pNode = pNewNode;
            for (size_t i = 0; i < newNodes.size(); i++) {
                newNodes[i]->release();
            }
        }
    } while (!done);

    int32_t newScore =
        pNode == pRootNode ?
        grid.m_pInfo->m_score :
        pNode->m_diff.m_pInfo->m_score;
    pNode->release();
    return newScore;
}

void solve(Grid& grid, vector<Action>* pTrace, bool bTest)
{
    srandom((unsigned int)time(0));
    ZoneMap zoneMap;
    makeZoneMap(&zoneMap, grid);

    vector<Pos> usefulCells;
    for (uint32_t y = 1; y <= grid.m_height; y++) {
        for (uint32_t x = 1; x <= grid.m_width; x++) {
            if (zoneMap.cell(x, y) >= 0) {
                usefulCells.push_back(Pos(x, y));
            }
        }
    }
    uint32_t numUsefulCells = (uint32_t)usefulCells.size();

    Node* pRootNode = new Node;

    Best best;
    best.update(grid, pRootNode);

    if (bTest) {
        puts("\x1b[2J\x1b[H");
        writeGrid(stdout, grid, true);
    }

    Plan* planArr[16];
    int32_t scoreArr[16];

    for (uint32_t i = 0; i < 16; i++) {
        planArr[i] = new Plan;
        if (i == 0) {
            scoreArr[i] = tryPlan(grid, pRootNode, best, *planArr[i], bTest);
            planArr[i]->m_bestVisited.resize(planArr[i]->m_visited.size());
            std::copy(planArr[i]->m_visited.begin(),
                      planArr[i]->m_visited.end(),
                      planArr[i]->m_bestVisited.begin());
        } else {
            scoreArr[i] = scoreArr[0];
            planArr[i]->m_bestVisited.resize(planArr[0]->m_bestVisited.size());
            std::copy(planArr[0]->m_bestVisited.begin(),
                      planArr[0]->m_bestVisited.end(),
                      planArr[i]->m_bestVisited.begin());
        }
    }

    while (true) {
        if (receivedSignal()) { break; }

        for (uint32_t i = 0; i < 15; i++) {
            if (scoreArr[i] <= scoreArr[i+1]) {
                std::swap(scoreArr[i], scoreArr[i+1]);
                std::swap(planArr[i], planArr[i+1]);
            }
        }

        uint32_t iPlan = 0;
        {
            for (uint32_t i = 1; i <= 15; i++) {
                if ((uint32_t)random() % 4 != 0) {
                    iPlan++;
                } else {
                    break;
                }
            }
        }
        Plan& plan = *planArr[iPlan];
        int32_t& score = scoreArr[iPlan];

        if (bTest) {
            //printf("iPlan = %u\n", iPlan);
        }
        bool didSomething = false;

#if 0
        for (uint32_t y = 1; y <= grid.m_height; y++) {
            if (receivedSignal()) { break; }
            if (didSomething) { break; }
            for (uint32_t x = 1; x <= grid.m_width; x++) {
                if (receivedSignal()) { break; }
                if (didSomething) { break; }
                if (zoneMap.cell(x, y) < 0) { continue; }

                bool toAdd =
                    plan.m_set.find(Pos(x, y)) == plan.m_set.end();

                if (toAdd) {
                    plan.m_set.insert(Pos(x, y));
                } else {
                    plan.m_set.erase(Pos(x, y));
                }

                int32_t newScore = tryPlan(grid, pRootNode, best, plan, bTest);
                if (newScore > score) {
                    score = newScore;
                    didSomething = true;
                } else if (newScore == score) {
                    if (toAdd) {
                        plan.m_set.erase(Pos(x, y));
                    } else {
                        didSomething = true;
                    }
                } else {
                    if (toAdd) {
                        plan.m_set.erase(Pos(x, y));
                    } else {
                        plan.m_set.insert(Pos(x, y));
                    }
                }
            }
        }
#endif

        uint32_t changeType = (uint32_t)random() % 15;
        if (bTest) {
            //printf("changeType = %u\n", changeType);
        }

        if (changeType < 5) {
            uint32_t numChanges = changeType + 1;
            vector<Pos> pos;
            pos.resize(numChanges);
            for (uint32_t i = 0; i < numChanges; i++) {
                uint32_t iPos = (uint32_t)random() % numUsefulCells;
                pos[i] = usefulCells[iPos];
                if (bTest) {
                    //printf("  (%u, %u)\n", pos[i].m_x, pos[i].m_y);
                }
                if (plan.m_set.find(pos[i]) == plan.m_set.end()) {
                    plan.m_set.insert(pos[i]);
                } else {
                    plan.m_set.erase(pos[i]);
                }
            }

            int32_t newScore = tryPlan(grid, pRootNode, best, plan, bTest);
            if (newScore > score) {
                score = newScore;
                didSomething = true;
                plan.m_bestVisited.resize(plan.m_visited.size());
                std::copy(plan.m_visited.begin(),
                          plan.m_visited.end(),
                          plan.m_bestVisited.begin());
            } else if (newScore == score &&
                       numChanges == 1 &&
                       (uint32_t)random() % 4 <
                       (plan.m_set.find(pos[0]) == plan.m_set.end() ?
                        3 : 1)) {
                didSomething = true;
                plan.m_bestVisited.resize(plan.m_visited.size());
                std::copy(plan.m_visited.begin(),
                          plan.m_visited.end(),
                          plan.m_bestVisited.begin());
            } else {
                for (uint32_t i = 0; i < numChanges; i++) {
                    if (plan.m_set.find(pos[i]) == plan.m_set.end()) {
                        plan.m_set.insert(pos[i]);
                    } else {
                        plan.m_set.erase(pos[i]);
                    }
                }
            }
            plan.m_visited.clear();
        } else if (changeType < 15) {
            uint32_t numChanges = (changeType%5) + 1;
            map<Pos, Pos>& theMap =
                changeType < 10 ? plan.m_parentMap : plan.m_nextMap;

            vector<Pos> pos;
            pos.resize(numChanges);
            vector<Pos> parent;
            parent.resize(numChanges);
            for (uint32_t i = 0; i < numChanges; i++) {
                if (!plan.m_bestVisited.empty() &&
                    (uint32_t)random() % 2 == 0) {
                    uint32_t iPos = (uint32_t)random() % (uint32_t)plan.m_bestVisited.size();
                    pos[i] = plan.m_bestVisited[iPos];
                    uint32_t iParent = (uint32_t)random() % (uint32_t)plan.m_bestVisited.size();
                    parent[i] = plan.m_bestVisited[iParent];
                } else {
                    uint32_t iPos = (uint32_t)random() % numUsefulCells;
                    pos[i] = usefulCells[iPos];
                    uint32_t iParent = (uint32_t)random() % numUsefulCells;
                    parent[i] = usefulCells[iParent];
                }
                if (theMap.find(pos[i]) == theMap.end()) {
                    theMap[pos[i]] = parent[i];
                    if (bTest) {
                        /*
                        printf("  (%u, %u) -> (%u, %u)\n",
                               pos[i].m_x, pos[i].m_y,
                               parent[i].m_x, parent[i].m_y);
                        */
                    }
                } else {
                    parent[i] = theMap[pos[i]];
                    theMap.erase(pos[i]);
                    if (bTest) {
                        /*
                        printf("  (%u, %u) -> nothing\n",
                               pos[i].m_x, pos[i].m_y);
                        */
                    }
                }
            }

            int32_t newScore = tryPlan(grid, pRootNode, best, plan, bTest);
            if (newScore > score) {
                score = newScore;
                didSomething = true;
                plan.m_bestVisited.resize(plan.m_visited.size());
                std::copy(plan.m_visited.begin(),
                          plan.m_visited.end(),
                          plan.m_bestVisited.begin());
            } else if (newScore == score &&
                       numChanges == 1 &&
                       theMap.find(pos[0]) == theMap.end()) {
                didSomething = true;
                plan.m_bestVisited.resize(plan.m_visited.size());
                std::copy(plan.m_visited.begin(),
                          plan.m_visited.end(),
                          plan.m_bestVisited.begin());
            } else {
                for (uint32_t i = 0; i < numChanges; i++) {
                    if (theMap.find(pos[i]) == theMap.end()) {
                        theMap[pos[i]] = parent[i];
                    } else {
                        theMap.erase(pos[i]);
                    }
                }
            }
            plan.m_visited.clear();
        }

        if (didSomething) { continue; }
    }

#if 1
    if (bTest)
    {
        Plan& plan = *planArr[0];
        set<Pos>::const_iterator iter = plan.m_set.begin();
        set<Pos>::const_iterator end = plan.m_set.end();
        for (; iter != end; ++iter) {
            printf("(%u, %u)\n", iter->m_x, iter->m_y);
        }
    }
    if (bTest)
    {
        Plan& plan = *planArr[0];
        map<Pos, Pos>::const_iterator iter = plan.m_parentMap.begin();
        map<Pos, Pos>::const_iterator end = plan.m_parentMap.end();
        for (; iter != end; ++iter) {
            printf("  (%u, %u) -> (%u, %u)\n",
                   iter->first.m_x, iter->first.m_y,
                   iter->second.m_x, iter->second.m_y);
        }
    }
    if (bTest)
    {
        Plan& plan = *planArr[0];
        map<Pos, Pos>::const_iterator iter = plan.m_nextMap.begin();
        map<Pos, Pos>::const_iterator end = plan.m_nextMap.end();
        for (; iter != end; ++iter) {
            printf("  ! (%u, %u) -> (%u, %u)\n",
                   iter->first.m_x, iter->first.m_y,
                   iter->second.m_x, iter->second.m_y);
        }
    }
#endif

    for (uint32_t i = 0; i < 16; i++) {
        if (bTest) {
            printf("Score %2u: %d\n", i, scoreArr[i]);
        }
        delete planArr[i];
    }

#if 0
    int32_t oldScore = best.m_score;
    do {
        if (receivedSignal()) { break; }
        oldScore = best.m_score;
        vector<Node*> lambdaNodes;
        search(grid, best.m_pNode, best, &lambdaNodes, bTest);
        if (bTest) {
            //printf("Best score: %d\n", best.m_score);
        }
    } while (best.m_score > oldScore);
#endif

    vector<Action> actions;
    Node* pCurNode = best.m_pNode;
    while (pCurNode->m_parent) {
        actions.push_back(pCurNode->m_action);
        pCurNode = pCurNode->m_parent;
    }

    pTrace->resize(actions.size());
    for (size_t i = 0; i < actions.size(); i++) {
        (*pTrace)[i] = actions[actions.size() - 1 - i];
    }

    if (best.m_state == State_Running) {
        pTrace->push_back(Action_Abort);
    }

    pRootNode->release();

    best.clear();
    if (bTest) {
        printf("Node count: %u\n", s_nodeCount);
    }
}
